# File Maker for Photo Color Adjustment 32/16/8bit Version
# Ver. 5.01c Mar. 2023 32bit image support version
# add Gamma adjustment function for mixing G image (4.52)
# add EventListner to Novice Dialog (4.53) 2022_08_06
# bug fix (4.54) 2022_08_20
# parameters modified (4.55)
# New Hybrid Algorism applyed and UI improved (4.56)
# add psedo flat-field correction algorism
# (4.57b) alternative test 2022_08_23
#    modify mixing G Gamma 1.15 > 1.56
#    modify green preserve mask
#    release (4.60a) 2022_09_09
#    add image condition option (4.60b)
#    modify FF correction sigma (4.60c)
#    modify user interface (4.61)
#    modify for preMixing (4.71a)
#    modify for green preserve mask (4.71b)
#    modify for green preserve mask parameter (4.71c)
#    support for 32 bit full color image (5.00)
#    modified New Hybrid algorithm (5.01) sigma value
#    bug fix for judgement of colorProcessor (5.01)
#    Change UI for Gamma value input (5.01)
#    bug fix for caling user made mask reading routine (5.01a)
#    change Premixng sigma factor (5.01a)
#    use premixed B image as PFFc applying image (5.01b)
#    make sigma value smaller for portrait-oriented image (5.01c)
# (c)2020-23 Ohnishi, Yasuo
from ij import IJ, ImagePlus
from ij import plugin
from ij import WindowManager as WM
from ij.io import OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter
from ij.plugin import Duplicator
from ij.process import ImageConverter, ByteProcessor
from ij.process import ImageProcessor
from ij.plugin.filter import ThresholdToSelection
from loci.plugins import BF
from java.awt.event import ItemListener, AdjustmentListener
from java.awt import Color
import os
import datetime
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')
def convValue(Value, Mode):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if Mode == "16-bit":
        Value = int(Value * 256)
    elif Mode == "32-bit":
        Value = Value / 255
    else:
        Value = int(Value)
    return Value
#-----------------
def set_Lmax(imp, Mode):
    # set Maximum value of Lightness
    L_max = imp.getProcessor().getMax()
    if int(L_max) > 65536:
        L_max = 16777215
    elif int(L_max) > 256 or Mode == "16-bit":
        L_max = 65535
    elif int(L_max) > 1.0 or Mode == "8-bit":
        L_max = 255
    else:
        L_max = 1.0
    return L_max
#-----------------
def convValue2(Value, L_max):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if L_max == 16777215:
        Value = int(Value * 65536)
    elif L_max == 65535:
        Value = int(Value * 256)
    elif L_max == 255:
        Value = int(Value)
    else: # L_max == 1.0
        Value = Value / 255
    return Value
#-----------------
def convValue1(Value, L_max):
#   Converting Luminance Value each in 8, 16, 32bit mode
    Value = int(Value * (256 / L_max))
    return Value
#-----------------
def bit32to16(imp, L_max):
    imp.getProcessor().multiply(65535 / L_max)
    imp.setDisplayRange(0, 65535)
    ImageConverter(imp).convertToGray16()
    return imp
#-----------------
def bit16to32(imp, L_max):
    ImageConverter(imp).convertToGray32()
    imp.getProcessor().multiply(L_max / 65535.0)
    imp.setDisplayRange(0, L_max)
    return imp
#-----------------
def convert1to65535(imp, L_max):
    imp.getProcessor().multiply(65535 / L_max)
    imp.setDisplayRange(0, 65535)
    return imp
#-----------------
def convert65535to1(imp, L_max):
    imp.getProcessor().multiply(L_max / 65535.0)
    imp.setDisplayRange(0, L_max)
    return imp
#-----------------
def imageBinary(impOrg, lowerV, upperV, roiColor, maskname, Mode, L_max):
    # make binary mask image with setThreshold
    # lowerV: lowerThreshold / upperV: upperThreshold (double)
    # roiColor: Color Value, ROI filled with (double)
    # maskname: name of mask image (string)
    imp = Duplicator().run(impOrg)
    if Mode == "32-bit":
        imp = bit32to16(imp, L_max)
        L_maxTmp = 65535
    else:
        L_maxTmp = L_max
    ip = imp.getProcessor()
    mask = ip.createProcessor(ip.width, ip.height)
    mask.setValue(convValue2(roiColor, L_maxTmp)) # with which color ROI filled
    ip.setThreshold(convValue2(lowerV, L_maxTmp), convValue2(upperV, L_maxTmp), \
        ImageProcessor.NO_LUT_UPDATE)
    roi = ThresholdToSelection.run(imp)
    mask.setRoi(roi)
    mask.fill(roi)
    maskimp = ImagePlus(maskname, mask)
    if Mode == "32-bit":
        maskimp = bit16to32(maskimp, L_max)
    imp.close()
    return maskimp
#-----------------
def clipUnderOver(imp, L_max):
    bit32to16(imp, L_max)
    bit16to32(imp, L_max)
    return imp
############### 2022/7/5
def make_Gaussian_imp(imp_org, sigma, Mode):
    imp_width = imp_org.getWidth()
    imp_height  = imp_org.getHeight()
    swidth = int(imp_width / 4)
    sheight = int(imp_height / 4)
    imp_g = Duplicator().run(imp_org)
    if Mode == "8-bit":
        ImageConverter(imp_g).convertToGray32()
        # imp_g will be gauss image of level adjusted imp_org
    imp_g.resize(swidth, sheight, 'bilinear')
    IJ.run(imp_g, "Gaussian Blur...", "sigma=" + str(sigma))
    imp_g.resize(imp_width, imp_height, 'bilinear')
    if Mode == "8-bit":
        ImageConverter(imp_g).convertToGray8()
    return imp_g # imp_g will be 16-bit
#----------------
def psedo_Flat_field_C(imp_g0, imp_b0, sigma, Mode, L_max, folder, \
        filename, parm_subtractBv, parm_leveladj, parm_correct_fac):
    """
        imp_g0: reference ch. imp, imp_b0: B ch. imp
        sigma: Gaussian blur sigma
        parm_subtractBv: Subtraction or not B when B is much higher than G
        parm_leveladj: adjust or not luminance level
        parm_correct_fac: manual correction factor
    """
### Options for developping
    if parm_subtractBv == False:
        # No Subtraction B when B is much higher than G
        test = "A"  # select test "A" or "B"
    else:  # Subtraction B when B is much higher than G
        test = "B"  # select test "A" or "B"
    writingout = False  # select writingout True or False
    if parm_leveladj != "No":
        meanadjustment = True # select mean adjustment True or False
    else:
        meanadjustment = False # select mean adjustment True or False
    ####
    imp_g1 = make_Gaussian_imp(imp_g0, sigma, Mode)
    imp_b1 = make_Gaussian_imp(imp_b0, sigma, Mode)
    """
        imp_g1: G ch. Gaussan blur image
        imp_b1: B ch. Gaussan blur image
    """
    if meanadjustment == True: # adjust level difference of G and B Ver. 3.53
        stat_g1 = imp_g1.getStatistics()
        stat_b1 = imp_b1.getStatistics()
        if parm_leveladj == "Auto":
            if stat_g1.mean < stat_b1.mean:
                # adjust level difference of G and B Gaussian image (Ver. 3.52)
                ### Bug Fixed (Ver. 4.54) 2022_08_20
                IJ.run(imp_g1, "Add...", "value=" + str(stat_b1.mean - stat_g1.mean))
                # imp_g1 must be lighter
            elif (test == "B") and (stat_g1.mean > stat_b1.mean): # modification Ver. 4.55
                IJ.run(imp_g1, "Subtract...", "value=" + str(stat_g1.mean - stat_b1.mean))
                # when subtract higher region of B than G and G is higher than B
                # also adjust level difference of G and B Gaussian image

        elif parm_leveladj == "Manual": # Manual adjustment of correction amount of B
            if stat_g1.mean < stat_b1.mean:
                # adjust level difference of G and B Gaussian image (Ver. 3.52)
                IJ.run(imp_g1, "Add...", "value=" + str(stat_b1.mean - stat_g1.mean))
            IJ.run(imp_g1, "Add...", "value=" + str(convValue2(parm_correct_fac, L_max)))
                # add correction value "parm_correct_fac"
                # modification in Ver. 4.55
                # make Manual value relative to auto adjusted value
    ic = plugin.ImageCalculator()
    imp_dif_gb = ic.run("Subtract create", imp_g1, imp_b1)
                # imp_dif_gb: darker region of Gaussian B than G
    imp_c_b0 = ic.run("Add create", imp_b0, imp_dif_gb)
                # image correction of B (subtract yellowed value from B)
    if Mode == "32-bit":
        imp_dif_gb = clipUnderOver(imp_dif_gb, L_max)
        imp_c_b0 = clipUnderOver(imp_c_b0, L_max)

    if test =="B":
        # subtract higher region of B than G
        imp_dif_bg = ic.run("Subtract create", imp_b1, imp_g1) # No window made
        imp_c_b = ic.run("Subtract create", imp_c_b0, imp_dif_bg) # No window made
        if Mode == "32-bit":
            imp_dif_bg = clipUnderOver(imp_dif_bg, L_max)
            imp_c_b = clipUnderOver(imp_c_b, L_max)
        imp_dif_bg.close()
    else:
        imp_c_b = imp_c_b0 # for test "A"
    stat_c_b = imp_c_b.getStatistics()
    """
    if stat_c_b.mean > stat_g.mean:
        IJ.run(imp_c_b, "Subtract...", "value=" + \
                        str(int(stat_c_b.mean - stat_g.mean)))
         not good!
    """
    if writingout == True:
        IJ.saveAsTiff(imp_dif_gb, folder + filename +\
                                    "_imp_Gaussian-dif_gb.tif")
        if test =="B":
            IJ.saveAsTiff(imp_dif_bg, folder + filename +\
                                    "_imp_Gaussian-dif_bg.tif")
    if Mode == "8-bit":
        ImageConverter(imp_c_b).convertToGray8()
    elif Mode == "16-bit":
        ImageConverter(imp_c_b).convertToGray16()
    elif Mode == "32-bit":
        ImageConverter(imp_c_b).convertToGray32()
    imp_dif_gb.close()
    imp_c_b0.close()
    return imp_c_b, imp_b1, imp_g1

#--------------------------
def make_imp_blueHigh(impRf2, impT, Mode, L_max):
    """
       making relatively high value region image of target channel
                                                    with user make mask
    """
    # read user made binary mask
    od = OpenDialog("Choose user made mask image file")
    folder = od.getDirectory()
    filename = od.getFileName()
    imps_Umask = BF.openImagePlus(folder + filename)
                    # Open file with Bio-format Plugin
    imp_binaryMask = imps_Umask[0] #imp0: Original Image
    bitdepth = imp_binaryMask.getBitDepth()
    if bitdepth == 32:
        Mask_L_max = set_Lmax(imp_binaryMask,"32-bit")
    if Mode == "8-bit":
        ImageConverter(imp_binaryMask).convertToGray8()
    elif Mode == "16-bit":
        ImageConverter(imp_binaryMask).convertToGray16()
        if bitdepth == 8:
            IJ.run(imp_binaryMask, "Multiply...","value=256")
    elif Mode == "32-bit":
        ImageConverter(imp_binaryMask).convertToGray32()
        if bitdepth != 32:
            imp_binaryMask.gerProcessor().Multiply(L_max / (2 ** bitdepth - 1))
    if bitdepth == 32:    # when Binary Mask is 32bit image
        imp_binaryMask.gerProcessor().Multiply(L_max / Mask_L_max)
    # combine mask
    ic = plugin.ImageCalculator()
    impT_high0 =  ic.run("Min create", impT, imp_binaryMask)
    impRf2_high0 =  ic.run("Min create", impRf2, imp_binaryMask)

    stat_impT_high0 = impT_high0.getStatistics()
    stat_impRf2_high0 = impRf2_high0.getStatistics()
    # input correction factor
    factor = blue_high_adjustment()
    #---test
    IJ.run(impRf2_high0, "Add...", \
            "value=" + str(stat_impT_high0.stdDev * factor))
    #---test end
    if stat_impRf2_high0.mean < stat_impT_high0.mean:
                # adjust Rf ch. value
        IJ.run(impRf2_high0, "Add...", \
            "value=" + str(stat_impT_high0.mean - stat_impRf2_high0.mean))
        impRf2_high1 =  ic.run("Min create", impRf2_high0, imp_binaryMask)
        impRf2_high0.close()
        impRf_highlight = impRf2_high1
    else:
        impRf_highlight = impRf2_high0
    impT_highlight  = impT_high0
    return impT_highlight, impRf_highlight

def blue_high_adjustment():
    gd = GenericDialog("Input correction adjustment parameters for bluer mask")
    gd.addSlider("Correction adjustment factor (%) x std Dev",\
                                                        0, 300, 100)
    gd.showDialog()
    bluer_factor = gd.getNextNumber()/100
    return bluer_factor


#---mask for excluding more bluer region from correction (for Hybrid correction)
def make_fg_mask1(impG, impB, folder, filename, inverse):
    """
    This routine is making mask for foreground layer
    """
    debugmode = False
    ic = plugin.ImageCalculator()
    imp8_B = Duplicator().run(impB) # corrected B ch.
    imp8_G = Duplicator().run(impG) # G ch.
    statB = impB.getStatistics()
    statG = impG.getStatistics()
    if statG.mean < statB.mean: #adjust B value relative to G
        IJ.run(imp8_G, "Add...", "value=" + str(statB.mean - statG.mean))
                                            # del int 4.82b
    imp8_B_G = ic.run("Subtract create", impB, imp8_G)
    imp8_G.close()
        # extract more green region
    stat = imp8_B_G.getStatistics()
    IJ.run(imp8_B_G, "Multiply...", "value=7")
    if inverse == True:
        IJ.run(imp8_B_G, "Invert", "") # bluer region  > blacker (opaque)
                                       # not to apply correction in bluer region
        if imp8_B_G.getBitDepth() == 32:
            imp8_B_G.getProcessor().add(L_max)
    if debugmode == True:
        IJ.saveAsTiff(imp8_B_G, folder + filename +"_ForeGrd_Mask_imp_B_G.tif")
        IJ.showMessage("imp8_B_G Max: " + str(stat.max / 256) + " Mean: "\
                                                + str(stat.mean / 256))
    imp8_0 = imp8_B_G
    if imp8_0.getBitDepth() == 32:
        imp8_0 = clipUnderOver(imp8_0, L_max)
    IJ.run(imp8_0, "Gamma...", "value=0.5")
    IJ.run(imp8_0, "Gaussian Blur...", "sigma=1")
    imp8_B.close()
    imp8_G.close()
    return imp8_0
    """
    imp8_0:
        region bluer > more opaque (not to apply correction)
    """
#---mask for preserving plants green region from correction
#                                                (for psedo FF correction)
def make_green_prsv_mask_new(impG, imp_c_b, impB, impR, Mode, L_max, \
                                            parm_Binary, folder, filename):
    """
    This routine is for preserving plants green option of foreground
                layer mask in Hybrid / psedo flat-field correction algorism
    impG: G ch. / imp_c_b: corrected B ch. / impB: original B ch. / impR: R ch.
    imp_R128: mask for background correction mask (R128) > but no use
    parm_Binary: imp_R128's threshold > but no use
    #!!!!!! imp_c_b may context adjusted B
    """
    debugmode = False
    ic = plugin.ImageCalculator()
    imp8_cB = Duplicator().run(imp_c_b) # corrected B ch.
    imp8_G = Duplicator().run(impG) # G ch.
    imp8_R = Duplicator().run(impR) # R ch.
    imp8_B130 = Duplicator().run(impB) # R ch.
    imp8_Brestriction = Duplicator().run(impB) # R ch.
    # making exclucing region image
    Threshold = parm_Binary # changed in 4.82
    imp8_B130 = imageBinary(impB, 0.0, Threshold, \
        255.999, "imp8_B130", Mode, L_max)
    imp8_B130.setDisplayRange(0.0, L_max)
        # ROI (under 130) will be filled with white
        # under 130 > white and over 130 > black

    # making relative green mask process start
    stat8_cB = imp8_cB.getStatistics()
    statG = impG.getStatistics()
    statR = impR.getStatistics()
    #if statG.mean < stat8_cB.mean: #adjust B value relative to G
    #    IJ.run(imp8_G, "Add...", "value=" + str(int(stat8_cB.mean - statG.mean)))
    stat8_G = imp8_G.getStatistics()
    if stat8_G.mean < statR.mean: #adjust R value relative to G
        IJ.run(imp8_G, "Add...", "value=" + str(statR.mean - stat8_G.mean))
                                                    # del int 4.82b
    #else:
    #    IJ.run(imp8_R, "Add...", "value=" + str(int(statG.mean - statR.mean)))
    IJ.run(imp8_R, "Multiply...", "value=0.20") # image R x 0.20 #*** 2023/1/5 revised
    IJ.run(imp8_cB, "Multiply...", "value=0.80") # image B x 0.8
    imp8_RandB = ic.run("Add create", imp8_R, imp8_cB)
        # imp8_RandB: R and B mixing image
    if imp8_RandB.getProcessor().getMin < 0.0:
        imp8_RandB = clipUnderOver(imp8_RandB, L_max)
    # G - (R x 0.2 + B x 0.8)
    IJ.run(imp8_G, "Add...", "value=" + str(convValue2(15, L_max)))
    imp8_G_cB = ic.run("Subtract create", imp8_G, imp8_RandB)
        # imp8_G_cB: relative G iamge (Greener = lighter)
    if imp8_G_cB.getBitDepth() == 32:
        imp8_G_cB = clipUnderOver(imp8_G_cB, L_max)
    imp8_G_cB.setDisplayRange(0.0, L_max)
        # data clip for 32bit image
    if debugmode == True:
        IJ.saveAsTiff(imp8_B130, folder + filename +"_imp8_B130.tif")
        IJ.saveAsTiff(imp8_G_cB, folder + filename +"_imp8_G_cB.tif")

    imp8_0 = ic.run("Min create", imp8_G_cB, imp8_B130)
        # excluding over 130 (Threshold) region
        # under 130: relative G image / over 130: black
    # deciding multiplying factor
    stat8_0 = imp8_0.getStatistics()
    mult_factor = convValue2(255.99, L_max) / stat8_0.max
    if debugmode == True:
        IJ.showMessage("stat8_0.max: " + str(stat8_0.max))
        IJ.showMessage("stat8_0.min: " + str(stat8_0.min))
        IJ.showMessage("mult_factor: " + str(mult_factor))
    mult_factor = mult_factor * 2.0
        #culculate multiplying factor
    IJ.run(imp8_0, "Multiply...", "value=" + str(mult_factor)) # mask x multiplying factor
    IJ.run(imp8_0, "Invert", "")
    if L_max == 1.0:
        #imp8_0.getProcessor().add(L_max)
        IJ.run(imp8_0, "Multiply...", "value=" + str(mult_factor))
    stat8_0 = imp8_0.getStatistics()
    if debugmode == True:
        IJ.showMessage("inverted stat8_0.max: " + str(stat8_0.max))
        IJ.showMessage("inverted stat8_0.min: " + str(stat8_0.min))
    if imp8_0.getBitDepth() == 32:
        imp8_0 = clipUnderOver(imp8_0, L_max)
    imp8_0.setDisplayRange(0.0, L_max)
        # more green region  > more black (opaque) and over region 130 > white
    # set restriction for green prsv mask with B value
    #restrictionFac = (255.99 - Threshold) / Threshold
    #IJ.run(imp8_Brestriction, "Multiply...", "value=" + str(restrictionFac))
    ImageConverter(imp8_Brestriction).convertToGray32()
    imp8_Brestriction.getProcessor().multiply(1.0 / convValue2(Threshold, L_max)) #*****
    if imp8_Brestriction.getProcessor().getMax > L_max:
        imp8_Brestriction = clipUnderOver(imp8_Brestriction, 1.0) # make imp8_Brestriction 0.0 to 1.0
    stat8_Br = imp8_Brestriction.getStatistics()
    if imp8_Brestriction.getProcessor().getMin < 0.0:
        imp8_Brestriction = clipUnderOver(imp8_Brestriction, L_max)
            # make imp8_Brestriction 0.0 to 1.0
    tmpMax = imp8_Brestriction.getProcessor().getMax()
    if debugmode == True:
        IJ.showMessage("imp8_Brestriction Max: " + str(tmpMax))
    ImageConverter(imp8_0).convertToGray32()
    tmpMax = imp8_0.getProcessor().getMax()
    if debugmode == True:
        IJ.showMessage("imp8_0 Max: " + str(tmpMax))
        IJ.saveAsTiff(imp8_0, folder + filename +"_imp8_0.tif")
        IJ.saveAsTiff(imp8_Brestriction, folder + filename +"_imp8_Brestriction.tif")

    imp8_1 = ic.run("Multiply create", imp8_0, imp8_Brestriction) # 32-bit
    if imp8_1.getBitDepth() == 32:
        imp8_1 = clipUnderOver(imp8_1, L_max)
        # make B higher region brighter
    IJ.run(imp8_1, "Gamma...", "value=1.5")
    tmpMax = imp8_1.getProcessor().getMax()
    if debugmode == True:
        IJ.showMessage("imp8_1 Max: " + str(tmpMax))

    mult_Factor2 = ((255 - Threshold) / Threshold) * 0.3
    # IJ.run(imp8_1, "Multiply...", "value=0.5")  # changed in 4.82
    imp8_1.getProcessor().multiply(mult_Factor2) # make brighter amount
    #imp8_2 = ic.run("Add create", imp8_0, imp8_1) # 32-bit
    imp8_2 = Duplicator().run(imp8_0) # 32-bit
    if imp8_2.getBitDepth() == 32:
        imp8_2 = clipUnderOver(imp8_2, L_max)
    imp8_2.setDisplayRange(0.0, L_max)

    imp8_cB.close()
    imp8_R.close()
    imp8_G_cB.close()
    imp8_B130.close()
    imp8_0.close()
    imp8_1.close()
    imp8_Brestriction.close()
    if Mode == "8-bit":
        ImageConverter(imp8_2).convertToGray8()
    elif Mode == "16-bit":
        ImageConverter(imp8_2).convertToGray16()
    return imp8_2
    """
    imp8_0:
        region relatively green > more opaque
    """

#-----Image Convert to Gray-------------
# 3 channel color image to mono channel grayscale image / Ver. 4.71
def imageConvert2Gray(imp, Mode):
    impsCh2 = ChannelSplitter.split(imp) #Channel Splitting of imp0
    ic = plugin.ImageCalculator()
    for i in impsCh2:
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
        if i == 0:
            IJ.run(i, "Multiply...", "value=0.299") # R
        elif i ==1:
            IJ.run(i, "Multiply...", "value=0.587") # G
        elif i ==2:
            IJ.run(i, "Multiply...", "value=0.114") # B
    impTmp0 = ic.run("Add create", impsCh2[0], impsCh2[1])
    impTmp1 = ic.run("Add create", impTmp0, impsCh2[2])
    impTmp0.close()
    return impTmp1

#---------------------------------
def preMixing(impT, impPrMix, parm_FFC_sigma, parm_correct_fac, Mode, \
                                                    L_max, folder, filename):
    # add parm_correct_fac Ver. 4.71
    """
        This routine is for premixing G channel to B for original B image
        impT: Target Ch.  impPrMix: Pre mixing Ch.
    """
    debugMode = False
    imp_c_Mx, imp_Gauss_preMix1, imp_Gauss_T1 = \
        psedo_Flat_field_C(impT, impPrMix, parm_FFC_sigma * 4, Mode, L_max, \
                            folder, filename, True, "Auto", parm_correct_fac)
    # if parm_FFC_sigma is larger, then image quality will be much better for pre mixing
    # make adjusted G ch.
    # imp_c_Mx: corrected Mixing ch.
    #   parm_subtractBv: True, parm_leveladj: "Auto"
    #***  IJ.run(imp_g1, "Add...", "value=" + str(convValue(parm_correct_fac, Mode)))
    impT_tmp03 = Duplicator().run(impT) # B ch.
    stat_impT = impT.getStatistics()
    stat_imp_c_Mx = imp_c_Mx.getStatistics()
    if stat_imp_c_Mx.mean > stat_impT.mean:
        IJ.run(imp_c_Mx, "Subtract...", "value=" + str(stat_imp_c_Mx.mean - stat_impT.mean))
    elif stat_imp_c_Mx.mean < stat_impT.mean:
        IJ.run(imp_c_Mx, "Add...", "value=" + str(stat_impT.mean - stat_imp_c_Mx.mean))
    IJ.run(impT_tmp03, "Multiply...","value=0.3")
    IJ.run(imp_c_Mx, "Multiply...","value=0.7")
    impT.close()
    ic = plugin.ImageCalculator()
    impT = ic.run("Add create", impT_tmp03, imp_c_Mx)
    # output for bugfixing
    if debugMode == True:
        IJ.saveAsTiff(imp_c_Mx, folder + filename + "_prMixingR.tif")
    #--
    impT_tmp03.close()
    imp_c_Mx.close()
    imp_Gauss_preMix1.close()
    imp_Gauss_T1.close()

    return impT

#-----------------------------------------
def RGBmanipulation(folder,filename,parm, imps):
    debugMode = False
#  input Parameters
    if parm is not None:
 # if parm is not null object, then...
        Mode, parm_Binary, parm_Gain, parm_Shadow, Show, parm_Prph, parm_FGGain,\
            parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, parm_subtractBv,\
            parm_leveladj, parm_correct_fac, HybridC, Fg_mask_type, bgG_gamma,\
            parm_ReadMask, DoPreMixing = parm
        print(Mode, parm_Binary, parm_Gain, parm_Shadow, Show, parm_Prph, parm_FGGain)
#     Mode: Image bit depth 16-bit or 8-bit
#     Show: Show working windows or not
#     parm_Binary: Threshold of fore / background
#     parm_Gain: Lightness of Distance adjusting Layer
#     parm_Shadow: Threshold of Dark part adjustment image
#     parm_Prph: Threshold of Periphery adjustment image
#     parm_FGGain: Threshold of Foreground mask
#     parm_Gmixrate: G mixing rate for foreground image in EPFFC only mode
#     parm_subtractBv: execute value subtraction of B value higher region
#     parm_leveladj: execute level adjustment G and B and it's type: Auto/Manual/None
#     parm_correct_fac: Manual correction factor
#     HybridC: execute hybrid correction or not and it's type
#     bgG_Gamma: Gamma value of mixing G for background image
#     Fg_mask_type: Type of mask for foreground layer Type1 to 4 and None
#     parm_ReadMask: read user made mask True/False
#     DoPreMixing: Do Premixing G to B for original B image True/False
# Always Open file with Bio-format Plugin
    imp0 = imps[0] #imp0: Original Image
    #-- adjustment Sigma of Extended psedo FF correction width image pixel size (4.60c)
    imp0_width = imp0.getWidth()
    imp0_height =  imp0.getHeight()
    imp0_long = max(imp0_width, imp0_height)
    sigma_factor = 10 / ((imp0_long // 1000) + 1)
    parm_FFC_sigma = parm_FFC_sigma / sigma_factor
    #--
    if Mode == "Auto":
        bitdepth = imp0.getBitDepth()
        if bitdepth == 8:
            Mode = "8-bit"
        elif bitdepth ==16:
            Mode = "16-bit"
        elif bitdepth ==24: # colorProcessor (5.01)
            Mode = "8-bit"
        elif bitdepth ==32:
            Mode = "32-bit"
        else:
            Mode = "16-bit"
    L_max = set_Lmax(imp0, Mode)
    if Mode == "32-bit":
        IJ.showMessage("32-bit image Lightness Max Value: " + str(L_max))
        imp0.setDisplayRange(0.0, L_max)
    if Show == "Yes":
 # Windows Showing Mode
        imp0.show()
#------
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
## R ch.
    impR =  impsCh[0]
    IJ.saveAsTiff(impR, folder + filename +"_R.tif")
## G ch.
    impG =  impsCh[1]
    IJ.saveAsTiff(impG, folder + filename +"_G.tif")
## B ch. & B x 0.3
    impB = Duplicator().run(impsCh[2])
    if DoPreMixing == False:
        IJ.saveAsTiff(impB, folder + filename +"_B.tif")
    else: # execute premixing G to B
        imp_nB = preMixing(impB, impG, parm_FFC_sigma / 3, parm_correct_fac, \
                                            Mode, L_max, folder, filename)
        IJ.saveAsTiff(imp_nB, folder + filename +"_B.tif")
        IJ.saveAsTiff(impB, folder + filename +"_Borg.tif")
        impB.close()
        impB = imp_nB
        # replace impB with premixing & corrected B image

# Show Windows
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
#--- preparetion for making layer images / to make (R+G)/2 image
#         & periphery correction layer
# G ch. x 0.5
#        Make impG05  as G x 0.5 Image
    impG05 = Duplicator().run(impG)
    if Show == "Yes":
        impG05.show()
    IJ.run(impG05, "Multiply...", "value=0.5")
# R ch. x 0.5
#        Make impR05  as R x 0.5 Image
    impR05 = Duplicator().run(impR)
    if Show == "Yes":
        impR05.show()
    IJ.run(impR05, "Multiply...", "value=0.5")
# B x 0.5 + R x 0.5 -- make peripherry corrction layer
    ic = plugin.ImageCalculator()
    imp5 = ic.run("Add create", impR05, impG05) # No window made
    #-----save periphery correction layer (no psedo FF correction)
    if (parm_FFCorrection == False) and (HybridC == "Off"):
        # traditional algorism
        IJ.saveAsTiff(imp5, folder + filename +"_Periph_Adjst.tif")
    #-----save periphery correction layer (psedo FF correction)
    else: # Hybrid & EPFFc
        imp_periph, imp_Gauss_b1, imp_Gauss_rg1 = \
            psedo_Flat_field_C(imp5, impB, parm_FFC_sigma, Mode, L_max, \
                    folder, filename, True, parm_leveladj, parm_correct_fac)
           # do psedo FF correction with option subtract B higher region
           # reference image must be impG or imp5?
        IJ.saveAsTiff(imp_periph, folder + filename +"_Periph_Adjst.tif")
        imp_periph.close()
    if Show == "Yes": #Hide the image windows to close
        imp5.show()
        impR05.hide()
        impG05.hide()
    impR05.close()
    impG05.close()

#--- making foreground layer image
#  executing psedo Flat-Field correction

#   always execute psedo_Flat_field_C for making mask
    imp_c_b_tmp, imp_Gauss_b1, imp_Gauss_g1 = \
        psedo_Flat_field_C(impG, impB, parm_FFC_sigma, Mode, L_max, folder, \
            filename, parm_subtractBv, parm_leveladj, parm_correct_fac)
#--   make blue higer image and combine imp_c_b
    if parm_ReadMask == True:
        impT = Duplicator().run(impB) # corrected B ch.
        impRf2 = Duplicator().run(impG) # G ch.

        impT_highlight, impRf_highlight = \
                make_imp_blueHigh(impRf2, impT, Mode, L_max)
        # make bluer region correction
        imp_c_b_high, imp_Gauss_b1_high, imp_Gauss_g1_high = \
            psedo_Flat_field_C(impRf_highlight, impT_highlight, \
            parm_FFC_sigma, Mode, L_max, folder, filename, parm_subtractBv,\
            "No", 0.0)
            # option with parm_leveladj: "BlueAdjustment"
        ic = plugin.ImageCalculator()
        imp_c_b =  ic.run("Max create", imp_c_b_tmp, imp_c_b_high)
        imp_c_b_tmp.close()
        imp_c_b_high.close()
        impT.close()
        impRf2.close()
    else:
        imp_c_b = imp_c_b_tmp
#--
##      Make impB03 as B x 0.3 Image > B:G ratio variable
    if parm_FFCorrection == True: # EPFFc only
        impB03 = Duplicator().run(imp_c_b)
            # use extended psedo FFc image
        Bmixrate = 1.0 - parm_Gmixrate # B:G mixing rate variable
        strValueB =  "value=" + str(Bmixrate)
        strValueG =  "value=" + str(parm_Gmixrate)
    elif (HybridC == "On") or (HybridC == "New"):
                    # Hybrid & New Hybrid algorithm
        impB03 = Duplicator().run(imp_c_b)
            # use extended psedo FFc image
        strValueB =  "value=0.3" # B:G mixing rate fixed
        strValueG =  "value=0.7"
    else: # existing algorithm
        impB03 = Duplicator().run(impB) # not to use FFcorection image
        strValueB =  "value=0.3"
        strValueG =  "value=0.7"
    statB03 = impB03.getStatistics()
            # use normal B image
    IJ.run(impB03, "Multiply...", strValueB)
    if Show == "Yes":
        impB03.show()
# G ch. x 0.7
##        Make impG07 as G x 0.7 Image
#    -----Next version modification will be here!
    if HybridC != "New":
        impG07 = Duplicator().run(impG)
    else: # HybridC == "New"
        imp_c_g, imp_Gauss_g2, imp_Gauss_b2 = \
            psedo_Flat_field_C(imp_c_b, impG, parm_FFC_sigma * 0.5, Mode, \
            L_max, folder, filename, "True", "Auto", 1.0)
            # modified in ver. 5.01
            # parm_subtractBv: True / parm_leveladj: Auto /
            # parm_correct_fac 1.0 / parm_FFC_sigma: 5.0 (in HybridC == "New")

        impG07 = imp_c_g
        imp_Gauss_g2.close()
        imp_Gauss_b2.close()

    statG07 = impG07.getStatistics()
    if statG07.mean < statB03.mean:
        IJ.run(impG07, "Add...", "value=" + str(statB03.mean - statG07.mean))
        # adjust G value for mixing to B
    IJ.run(impG07, "Multiply...", strValueG)
    if Show == "Yes":
        impG07.show()
# B x 0.3 + G x 0.7
#---  making foreground layer image & save
    ic = plugin.ImageCalculator()
    imp3 = ic.run("Add create", impB03, impG07) # No window made
    IJ.saveAsTiff(imp3, folder + filename +"_B_Fg.tif")
        # even in new alogorithm G mixing is needed to improve texture
    imp3.close()
#--- making background layer image
# (G ch.+30) x 0.7
##        Make impG30x07 as (G+30) x 0.7 Image
    if parm_FFCorrection == False:  # Applying Hybrid or Traditional algorism
        # Mixing lighter G image with B for background layer
        impG30x07 = Duplicator().run(impG)
        if Show == "Yes":
            impG30x07.show()
        # Gain for background layer image
        # appropriate gamma is /1.45(= 0.69) for perceptive TRC image
        if parm_FGGain == 128: # in case of perceptive gamma TRC image.
            IJ.run(impG30x07, "Gamma...", "value=" + str(bgG_gamma))
                                                    # originally 0.69
        else: # in case of linear image.
            IJ.run(impG30x07, "Gamma...", "value=" + str(bgG_gamma))
                                                    # originally 0.7
        IJ.run(impG30x07, "Multiply...", "value=0.7")
    ## B x 0.3 + (G+30) x 0.7 + x(parm_Gain)
        ic = plugin.ImageCalculator()
        imp4 = ic.run("Add create", impB03, impG30x07) # No window made
    #  add B x 0.3 + (G+30) x 0.7 + x
        IJ.run(imp4, "Add...", "value=" + str(convValue2(parm_Gain, L_max)))
                                        # add Value parm_Gain
        if Show == "Yes":
 #Hide the image windows to close
            impG30x07.hide()
        impG30x07.close()
    else:  # Apply Flat-field correction (exclude hybird)
        # No G ch. mixing in background layer
        imp4 = Duplicator().run(imp_c_b)
    IJ.saveAsTiff(imp4, folder + filename +"_B_Bg.tif")
    #imp4.close()
    if Show == "Yes":
        impG07.hide()
        impB03.hide()
    impG07.close()
    impB03.close()
#--- making Periphery Adjustment Mask
# Periphery Adjustment Mask: B - ((R+G)/2-10)
    impGR = Duplicator().run(imp5) #imp5: (G+R)/2 image
    if Show == "Yes":
        impGR.show()
    IJ.run(impGR, "Subtract...", "value="+ str(convValue2(parm_Prph, L_max)))
                                            # subtract Value parm_Prph /10
    if Mode == "32-bit":
        impGR = clipUnderOver(impGR, L_max)

    ic = plugin.ImageCalculator()
    if parm_FFCorrection == False: # traditional & Hybrid algorism
        imp2GR = ic.run("Subtract create", impB, impGR)
                                    #impB - impGR:(G+R/2)
    else: # EPFFc only algorism
        imp2GR = ic.run("Subtract create", impB, impGR)
                            #corrected impB - impGR:(G+R/2)
    if Show == "Yes":
        imp2GR.show()
    IJ.run(imp2GR, "Multiply...", "value=4") # Save the Subtracted image x 4
    if Mode == "32-bit":
        imp2GR = clipUnderOver(imp2GR, L_max)
    IJ.saveAsTiff(imp2GR, folder + filename +"_Periph_Adjst_Mask.tif")
    if Show == "Yes":
 #Hide the image windows to close
        imp2GR.hide()
        impGR.hide()
    imp2GR.close()
 #Close the images
    impGR.close()
#
#--- making dark parts correction layer
# Dark Parts 0 - parm_Shadow
    imp6 = Duplicator().run(imp5) #imp5: (G+R)/2 image, imp6: dark part image
    imp6.setTitle("imp6")
    if Show == "Yes":
        imp6.show()
    if Mode == "32-bit":
        imp6 = bit32to16(imp6, L_max)
        L_maxTmp =65535
    else:
        L_maxTmp = L_max
    IJ.run(imp6, "Add...", "value=" + \
        str(convValue2(255.99 - parm_Shadow, L_maxTmp)))
    IJ.run(imp6, "Subtract...", "value=" + \
        str(convValue2(255.99 - parm_Shadow, L_maxTmp)))
    if Mode == "32-bit":
        imp6 = bit16to32(imp6, L_max)
    imp_dark = imp6
    IJ.saveAsTiff(imp_dark, folder + filename +"_Dark.tif")
    if Show == "Yes":
 #Hide the image windows to close
        imp5.hide()
        imp6.hide()
    imp5.close()
 #Close the images
    imp6.close()
#--- making background correction mask
# Make Binary Mask: Threshold R parm_Binary
    imp9_0 = imageBinary(impR, parm_Binary, \
        255.99, 255.999, "imp9_0", Mode, L_max)
                            # Make binary Mask: black for under threshold
                            # ROI (under 130) will be filled with white
    ic = plugin.ImageCalculator()
    imp9a = make_fg_mask1(imp4, impB, folder, filename, True)
        # making bluer imege [compare background image (imp4) with impB]
        # inverse: True
    imp4.close()
    imp9 = ic.run("Min create", imp9_0, imp9a)
        # imp9: under mid point region + bluer region reservation mask
    if Show == "Yes":
        imp9.show()
    imp9_0.close()
    IJ.saveAsTiff(imp9, folder + filename +"_Bg_Thld_Mask.tif")

#--- making foreground correction mask
# R + parm_FGGain(128) ## mask image for foreground correction layer
#---- traditional R+128 mask
    if Fg_mask_type == "Type1":  # Tradisional
        imp8 = Duplicator().run(impR)
        IJ.run(imp8, "Add...", "value=" + str(convValue2(parm_FGGain, L_max)))
        if Show == "Yes":
            imp8.show()
        IJ.saveAsTiff(imp8, folder + filename +"_Fg_Mask.tif")
        if Show == "Yes":
            imp8.hide()
        imp8.close()
#---- mask for applying correction in region where impB is darker than impG
#                                                         (in Hybrid Correction)
    elif (Fg_mask_type == "Type2") or (Fg_mask_type == "Type4"):
                                                # applying Hybirid correction
        ic = plugin.ImageCalculator()
        imp8a = make_fg_mask1(impG, impB, folder, filename, True)
                                                    # inverse: True
        if Fg_mask_type == "Type4": # with green preserve mask
            IJ.saveAsTiff(imp8a, folder + filename +"_Fg_Mask(Type2).tif")
            imp8b = make_green_prsv_mask_new (impG, impB, impB, impR, \
                Mode, L_max, parm_Binary, folder, filename) # change 4.71
                # imp_nB: preMixed impB
            sigma_factor = 10 / ((imp0_width // 1000) + 1)
            sigma = 5 / sigma_factor
            IJ.run(imp8b, "Gaussian Blur...", "sigma=" + str(sigma))
            imp8 = ic.run("Min create", imp8a, imp8b)
        else: # Type2
            imp8 = imp8a
        # IJ.run(imp8, "Multiply...", "value=" + str(convValue(1, Mode)))
                                        # 16, 32bit image > Multiply 256 ^ 1 or 2
        imp8.updateAndDraw()
#---- no mask for foreground layer
    elif Fg_mask_type == "None":
        ip = impR.getProcessor()
        blank_mask = ByteProcessor(ip.width, ip.height)
        blank_mask.BLACK_AND_WHITE_LUT
        blank_mask.setValue(255)
        imp8 = ImagePlus("foreground layer mask", blank_mask)
        if Mode == "16-bit":
                ImageConverter(imp8).convertToGray16()
        IJ.run(imp8, "Add...","value=" + str(convValue2(255, L_max)))
#---- mask for adjustment plants green Type3
    else:   # Type3
        imp8 = make_green_prsv_mask_new\
            (impG, impB, impB, impR, Mode, L_max, parm_Binary, folder, filename)
        sigma_factor = 10 / ((imp0_width // 1000) + 1)
        sigma = 5 / sigma_factor
        IJ.run(imp8, "Gaussian Blur...", "sigma=" + str(sigma))

            # change 4.71
            # imp_nB: preMixed impB

    if Show == "Yes":
        imp8.show()
    if debugMode == True:
        IJ.showMessage("Mask Type: " + Fg_mask_type)
    IJ.saveAsTiff(imp8, folder + filename +"_Fg_Mask.tif")
    if Show == "Yes":
        imp8.hide()
    imp8.close()
    imp_c_b.close()
    imp_Gauss_b1.close()
    imp_Gauss_g1.close()
#---------------------
    if Show == "Yes":
 #Hide the imp9 window to close
        imp9.hide()
        impB.hide()
    imp0.close() # 4.71
    imp9.close()
    impR.close()
    impG.close()
    impB.close()
#------------------------------------------------
class DialogPreviewer(ItemListener, AdjustmentListener):
    def __init__(self, choice, slider, nfield,  nfield1, nfield3, nfield2=None):
        """
          ##choice-TRC > parameter fields
          ##choice-Hybrid correction (novice)> parameter fields
            choice: a java.awt.Choice
            processtype: string
            nfield:  a java.awt.TextField
            slider:  a java.awt.Scrollbar

        """
        self.slider = slider # Mixing G ch. Gamma
        self.nfield = nfield # Background mask threshold
        self.nfield1 = nfield1 # Shadow area threshold
        self.nfield2 = nfield2 # Periphery threshold
        self.nfield3 = nfield3 # slider numeric field
        # default: self.nfield3.setText("1.56")
        # default: self.slider.setValue(156) # 1.56
        self.choice1 = choice
        #IJ.showMessage("Value set: " + str(self.slider.getValue()))

#        self.processtype = processtype

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        choice_item  = self.choice1.getSelectedItem()
        print "itemStateChanged", str(choice_item)
        if choice_item == "Perceptual": # Select TRC
            self.nfield.setText("128")
            self.nfield1.setText("25")
            if self.nfield2 != None:
                self.nfield2.setText("10")
                self.nfield2.setBackground(Color.white)
            self.nfield.setBackground(Color.white)
            self.nfield1.setBackground(Color.white)
        elif choice_item == "Linear": # Select TRC
            self.nfield.setText("56")
            self.nfield1.setText("3")
            if self.nfield2 != None:
                self.nfield2.setText("0")
                self.nfield2.setBackground(Color.cyan)
            self.nfield.setBackground(Color.cyan)
            self.nfield1.setBackground(Color.cyan)
        elif choice_item == "Off": # Select Hypbid Correction
            self.nfield3.setText("1.45")
            self.slider.setValue(145) #1.46
            self.nfield3.setBackground(Color.pink)
        elif choice_item == "On" or choice_item == "New Method On" :
                                                # Select Hypbid Correction
            self.nfield3.setText("1.56")
            #IJ.showMessage("Value set: " + str(self.slider.getValue()))
            self.slider.setValue(156) # 1.56
            self.nfield3.setBackground(Color.white)

    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on the state of the scroll bar. """
        value = self.nfield3.getText()
        if value == "1.00":
            self.nfield3.setBackground(Color.lightGray)
        elif value == "1.45":
            self.nfield3.setBackground(Color.pink)
        elif value == "1.56":
            self.nfield3.setBackground(Color.white)
        else:
            self.nfield3.setBackground(Color.yellow)
        return
#------------------------------------------------
class DialogPreviewer1 (ItemListener):
    def __init__(self, choice, choice2, slider, nfield,  nfield1, \
                                        nfield3, nfield4, nfield5, nfield6):
        """
           ##choice-Hybrid correction (expert)> parameter fields
             choice: a java.awt.Choice
             processtype: string
             nfield:  a java.awt.TextField
             slider:  a java.awt.Scrollbar
        """
        self.slider = slider # Mixing G ch. Gamma
        self.nfield = nfield # Background mask threshold
        self.nfield1 = nfield1 # Shadow area threshold
        self.nfield3 = nfield3 # slider numeric field
        self.choice1 = choice  # Hybrid correction
        self.nfield4 = nfield4 # Sigma
        self.nfield5 = nfield5 # G mix ratio
        self.nfield6 = nfield6 # incremental amount
        self.choice2 = choice2 # incremental amount on/off
        self.choice_item2  = self.choice2.getSelectedItem()

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        choice_item  = self.choice1.getSelectedItem()
        print "itemStateChanged", str(choice_item)
        if choice_item == "G Mixing":
                                        # Select Traditional algorism
            self.nfield3.setText("1.45")
            self.slider.setValue(145) #1.46
            self.nfield3.setBackground(Color.pink)
            self.nfield4.setBackground(Color.gray) # Sigma
            self.nfield5.setBackground(Color.gray) # G mix ratio
            self.nfield6.setBackground(Color.gray) # incremental amount
        elif choice_item == "Hybrid" or choice_item == "New Hybrid" :
                                            # Select Hypbid Correction
            self.nfield3.setText("1.56")
            self.slider.setValue(156) # 1.56
            self.nfield3.setBackground(Color.white)
            self.nfield4.setBackground(Color.green) # Sigma
            if choice_item == "New Hybrid":
                self.nfield4.setText("5.0")
            self.nfield5.setBackground(Color.gray) # G mix ratio
            if self.choice_item2  == "Manual":
                self.nfield6.setBackground(Color.green)
            #self.checkbox.setState(False)
        elif choice_item == "EPFFc":
                        # Select EPFFc
            self.nfield3.setText("1.00")
            self.slider.setValue(100) # 1.00
            self.nfield3.setBackground(Color.lightGray)
            self.nfield4.setBackground(Color.green) # Sigma
            self.nfield5.setBackground(Color.green) # G mix ratio
            if self.choice_item2  == "Manual":
                self.nfield6.setBackground(Color.green)

            #self.checkbox.setState(False)

        return
#------------------------------------------------
class DialogPreviewer2(ItemListener):
    def __init__(self, choice, slider, nfield3, nfield4, nfield5, \
                                    nfield6, choice2, choice3, checkbox2):
        """
           ##checkbox-pFF correction > parameter fields
             choice: a java.awt.Choice
             checkbox:  a java.awt.Checkbox
        """
        self.slider = slider # Mixing G ch. Gamma
        self.choice1 = choice # Hybrid c on/off
        choice_item1  = self.choice1.getSelectedItem()
        # self.checkbox = checkbox # apply FF correction
        self.nfield3 = nfield3 # slider's numeric field
        self.nfield4 = nfield4 # Sigma
        self.nfield5 = nfield5 # G mix ratio
        self.nfield6 = nfield6 # incremental amount
        #check_state  = checkbox.getState()  # apply FF correction
        self.choice2 = choice2 # incremental amount on/off
        choice_item2  = self.choice2.getSelectedItem()
                            # adjustment amount auto / manual / none
        self.choice3 = choice3 # imege conditon
        choice_item3  = self.choice3.getSelectedItem()
        self.checkbox2 = checkbox2 # Apply pre-mixing to B
        if choice_item1 == "EPFFc": # apply FF correction only
            self.nfield4.setBackground(Color.green)
            self.nfield5.setBackground(Color.green)
            if choice_item2  == "Manual":
                self.nfield6.setBackground(Color.green)
            else:
                self.nfield6.setBackground(Color.gray)
        elif  choice_item1 == "Hybrid" or  choice_item1 == "New Hybrid":
                                                            # apply Hybrid
            self.nfield4.setBackground(Color.green)
            self.nfield5.setBackground(Color.gray)
            if choice_item2  == "Manual":
                self.nfield6.setBackground(Color.green)
            else:
                self.nfield6.setBackground(Color.gray)
        else:  # G mixing only
            self.nfield4.setBackground(Color.gray)
            self.nfield5.setBackground(Color.gray)
            self.nfield6.setBackground(Color.gray)
        if nfield3.getText() == "1.00": # if slider value is 100
            self.nfield3.setBackground(Color.lightGray)
        elif nfield3.getText() == "1.45": # if slider value is 145
            self.nfield3.setBackground(Color.pink)
        elif nfield3.getText() == "1.56": # if slider value is 145
            self.nfield3.setBackground(Color.white)
        else:
            self.nfield3.setBackground(Color.yellow)
        if choice_item3 == "Normal": # image conditionp
            self.nfield4.setText("10.0") # set Sigma value
            self.checkbox2.setState(False)
        else: # "Coarse"
            choice_item1  = self.choice1.getSelectedItem()
            if choice_item1 != "G Mixing":
                self.nfield4.setText("5.0")

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        choice_item1  = self.choice1.getSelectedItem()
                                        # algorism selection
        choice_item2  = self.choice2.getSelectedItem()
                                    # adjustment amount auto / manual / none
        choice_item3  = self.choice3.getSelectedItem()
                                         # image condition normal / coarse

        if choice_item2 == "Manual": # psedo FF correction On
            self.nfield6.setBackground(Color.green)
        else:
            self.nfield6.setBackground(Color.gray)
        if choice_item3 == "Normal": # image conditionp
            if choice_item1 != "New Hybrid":
                self.nfield4.setText("10.0") # set Sigma value
            self.checkbox2.setState(False)
        else: # "Coarse"
            choice_item1  = self.choice1.getSelectedItem()
            if choice_item1 != "G Mixing":
                self.nfield4.setText("5.0")
            self.checkbox2.setState(True)
        return

#------------------------------------------------
# Novice Mode: Get parameters from the input DialogBox
def NoviceMenu(path, Vmean):
    gd = GenericDialog("Novice Menu: Enter Parameters")
    TRCType = ["Perceptual", "Linear"]
    gd.addChoice("Image TRC Type", TRCType, TRCType[0])
    if Vmean < 128.0:
        BkThs = 128.0
    else:
        BkThs = int(Vmean)
    gd.addNumericField("Background Mask Threshold (0 to 255)", int(BkThs) , 0)
    # gd.addNumericField("Background Image Gain (Max to 100)", 0, 0)
    gd.addSlider("Background mixing G ch. Gamma (increase > bluer)",\
                                                    1.00, 1.70, 1.56 ,0.01)
    gd.addNumericField("Shadow Area Threshold (0 to 255)", 25, 0)
    HybridC_Choice = ["Off", "On", "New Method On"]
    gd.addChoice("Apply hybrid correction?", \
                                        HybridC_Choice, HybridC_Choice[1])
    gd.addCheckbox("(in Hybrid correction) Preserve plants green", False)
    condition_Choice = ["Normal","Coarse"]
    gd.addChoice("Image condition", condition_Choice, condition_Choice[0])
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])
    gd.addMessage("If you would like to use Expert Menu,"+\
                                            " Please cancel this menu.")
    gd.addMessage("Recommended Parameters for Linear Images\n"+ \
        " 56 for Background Mask Threshold\n 3 for Shadow Area Threshold\n" +\
        " 0 for Periphery Adjst Threshold")

    # The UI elements for the above inputs
    choice = gd.getChoices().get(0) # TRC Type
    nfield = gd.getNumericFields().get(0) # Background mask threshold
    slider = gd.getSliders().get(0) # Mixing G Gamma
    nfield3 = gd.getNumericFields().get(1) # slider numeric field
    # Attention! Sliders has also NumericFields, so attention
    #                                   to index number for NFields!
    nfield1 = gd.getNumericFields().get(2) # Shadow area threshold
    choice1 = gd.getChoices().get(1) # Hybrid correction on/off
    # Set event previewer
    previewer = DialogPreviewer(choice, slider, nfield, \
                                        nfield1, nfield3)
    previewer1 = DialogPreviewer(choice1, slider, nfield, \
                                        nfield1, nfield3)
    choice.addItemListener(previewer)
    slider.addAdjustmentListener(previewer)  # Slider
    choice1.addItemListener(previewer1)

    gd.showDialog()

    # get values
    parm_Mode   = "Auto"
    TRCselected   = gd.getNextChoice()
    if TRCselected == "Perceptual":
        parm_FGGain = 128
        parm_Prph = 10
    else:
        parm_FGGain = 56
        parm_Prph = 0
    parm_Binary   = gd.getNextNumber() # Fg / Bg threshold
    parm_Gain = gd.getNextNumber() # (5.01a) Gamma value for Bg
    parm_Shadow = gd.getNextNumber() # Shadow correction threshold
    parm_HybridC   = gd.getNextChoice() # Correction method
    if parm_HybridC == "New Method On":
        parm_HybridC = "New"
    green_preserve = gd.getNextBoolean() # apply green preserve True / False
    parm_Condition   = gd.getNextChoice() # Image condition Normal / Coarse
    parm_Show   = gd.getNextChoice() # Show windows?
    # following parameters are fixed in Novice Mode
    parm_FFCorrection = False # apply only psedo flat-field correction > false
    if parm_Condition == "Normal" and parm_HybridC !="New":
        parm_FFC_sigma = 10.0
        DoPremixing = False
    else: # image condition: "Coarse" or parm_HybridC: "New"
        parm_FFC_sigma = 5.0
        DoPremixing = True
    parm_Gmixrate = 0.0
    parm_leveladj  = "Auto" # apply level adjustment between channels
    parm_correct_fac = 0.0 # correction adjustment factor
    parm_subtractBv = False
    bgG_gamma = 1.0 / parm_Gain # inverse value of gamma (actual applying power)
    parm_Gain = 0.0
    if parm_Binary > 255: #Set Max Value
        parm_Binary = 255
    if parm_Gain > 1.7:
        parm_Gain = 1.7
    if parm_Shadow > 255: #Set Max Value
        parm_Shadow = 255
    if green_preserve == True: #changed in 4.82
        Fg_mask_type = "Type4"
    elif green_preserve == False:
        Fg_mask_type = "Type2"
    parm_ReadMask = False
    if gd.wasCanceled():
        choice.removeItemListener(previewer)
        choice1.removeItemListener(previewer1)
        parm_Mode   = "Expert"
        IJ.log("Expert Menu called")
        return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow,\
            str(parm_Show), parm_Prph, parm_FGGain, parm_FFCorrection, \
            parm_FFC_sigma, parm_Gmixrate, parm_subtractBv, parm_leveladj, \
            parm_correct_fac, parm_HybridC, Fg_mask_type, bgG_gamma, \
            parm_ReadMask, DoPremixing
    write_out(path, parm_Binary, parm_Gain, parm_Shadow, parm_Prph, \
            parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, \
            Fg_mask_type, bgG_gamma, parm_Condition)
    choice.removeItemListener(previewer)
    choice1.removeItemListener(previewer1)
    return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow, \
        str(parm_Show), parm_Prph, parm_FGGain, parm_FFCorrection, \
        parm_FFC_sigma, parm_Gmixrate, parm_subtractBv, parm_leveladj, \
        parm_correct_fac, parm_HybridC, Fg_mask_type, bgG_gamma, \
        parm_ReadMask, DoPremixing
#----------------------------------------------------
# Expert Mode: Get parameters from the input DialogBox
def getParameters(path, Vmean):
    oldparms = read_in(path, Vmean)
    gd = GenericDialog("Expert Menu: Enter Parameters")
    TRCType = ["Perceptual", "Linear"]
    if int(oldparms["parm_FGGain"]) == 128:
        gd.addChoice("Image TRC Type", TRCType, TRCType[0])
    else:
        gd.addChoice("Image TRC Type", TRCType, TRCType[1])

    gd.addNumericField("Background Mask Threshold (0 to 255)",\
                                            float(oldparms["parm_Binary"]), 0)
    gd.addSlider("Background mixing G ch. Gamma (increase > bluer)", \
                            1.00, 1.70, 1.00 / float(oldparms["bgG_gamma"]), 0.01)
    gd.addNumericField("Shadow Area Threshold (0 to 255)", \
                                            float(oldparms["parm_Shadow"]), 0)
    gd.addNumericField("Periphery Adjst Threshold (-10 to 20)", \
                                            float(oldparms["parm_Prph"]), 0)

    gd.addMessage("                        in algorism /" + \
                    " EPFFc: Apply Extended Psedo Flat-Field Correction only")

    HybridC_Choice = ["G Mixing", "Hybrid", "New Hybrid", "EPFFc"]
    if float(oldparms["bgG_gamma"]) == 0.69:
        gd.addChoice("Select algorism", HybridC_Choice, HybridC_Choice[0])
    elif float(oldparms["bgG_gamma"]) == 1.00:
        gd.addChoice("Select algorism", HybridC_Choice, HybridC_Choice[3])
    else:
        gd.addChoice("Select algorism", HybridC_Choice, HybridC_Choice[1])
    mask_type = ["Type1", "Type2", "Type3", "Type4","None"]
    if oldparms["Fg_mask_type"] == "Type1":
        gd.addChoice("Mask type for foreground correction layer", \
                                            mask_type, mask_type[0])
    elif oldparms["Fg_mask_type"] == "Type3":
        gd.addChoice("Mask type for foreground correction layer", \
                                            mask_type, mask_type[2])
    elif oldparms["Fg_mask_type"] == "Type4":
        gd.addChoice("Mask type for foreground correction layer", \
                                            mask_type, mask_type[3])
    elif oldparms["Fg_mask_type"] == "None":
        gd.addChoice("Mask type for foreground correction layer", \
                                            mask_type, mask_type[4])
    else:
        gd.addChoice("Mask type for foreground correction layer", \
                                            mask_type, mask_type[1])
    gd.addMessage("       Type1: R+128 mask " + \
          "       Type2: Preserve Lighter B region " \
        + "\n       Type3: Preserve plants green        Type4: Type2 + 3")
    gd.addMessage("-----------------------------------")
    gd.addMessage("[Extended Psedo Flat-Field correction (EPFFc) parameters]")
    gd.addNumericField("Sigma for Flat-Field correction", \
                                    float(oldparms["parm_FFC_sigma"]), 1)
    gd.addNumericField("G mix rate to B (available in EPFFc only mode)", \
                                    float(oldparms["parm_Gmixrate"]), 1)
    leveladj = ["No","Auto","Manual"]
    if oldparms["parm_leveladj"] == "Auto":
        gd.addChoice("Correction Amount adjustment", leveladj, leveladj[1])
    elif oldparms["parm_leveladj"] == "Manual":
        gd.addChoice("Correction Amount adjustment", leveladj, leveladj[2])
    else:
        gd.addChoice("Correction Amount adjustment", leveladj, leveladj[0])
    gd.addNumericField("Incremental correcting amount (in Manual 0-255)", \
                                    float(oldparms["parm_correct_fac"]), 0)
    gd.addCheckbox("Subtract Pixel Value if B > G",\
                                        eval(oldparms["parm_subtractBv"]))
    gd.addCheckbox("Read user made bluer mask", False)
    gd.addCheckbox("Apply G texture info to original B ", True)
    gd.addMessage("-----------------------------------")
    condition_Choice = ["Normal","Coarse"]
    if oldparms["parm_Condition"] == "Normal":
        gd.addChoice("Image condition", condition_Choice, condition_Choice[0])
    elif oldparms["parm_Condition"] == "Coarse":
        gd.addChoice("Image condition", condition_Choice, condition_Choice[1])
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])

    # The UI elements for the above inputs
    choice = gd.getChoices().get(0) # TRC Type
    nfield = gd.getNumericFields().get(0) # Background mask threshold
    slider = gd.getSliders().get(0) # Mixing G Gamma
    nfield3 = gd.getNumericFields().get(1) # slider numeric field
    # Attention! Sliders has also NumericFields,
    #           so attention to index number for NFields!
    nfield1 = gd.getNumericFields().get(2) # Shadow area threshold
    nfield2 = gd.getNumericFields().get(3) # Periphery threshold
    choice1 = gd.getChoices().get(1) # Hybrid correction on/off
    choice3 = gd.getChoices().get(3) # increment adjustment
    choice4 = gd.getChoices().get(4) # image condition
    #checkbox = gd.getCheckboxes().get(0) # Apply FF correction
    nfield4 = gd.getNumericFields().get(4) # Sigma for FF correction
    nfield5 = gd.getNumericFields().get(5) # G mix ratio
    nfield6 = gd.getNumericFields().get(6) # Incremantarl correction amount
    checkbox2 = gd.getCheckboxes().get(2) # Apply pre mixing to B


    # Set event previewer
    #  TRC selection
    previewer = DialogPreviewer(choice, slider, nfield, \
                                        nfield1, nfield3, nfield2)
    #  Hybric correction
    previewer1 = DialogPreviewer1(choice1, choice3, slider, nfield, \
                                nfield1, nfield3, nfield4, nfield5, nfield6)
    #  Apply extended psedo FF correction
    previewer2 = DialogPreviewer2(choice1, slider, nfield3, \
                    nfield4, nfield5, nfield6, choice3, choice4, checkbox2)

    choice.addItemListener(previewer)  # TRC Type
    slider.addAdjustmentListener(previewer)  # Slider
    choice1.addItemListener(previewer1)  # Hybrid correction on/off
    # checkbox.addItemListener(previewer2)  # Apply FF correction
    choice3.addItemListener(previewer2)  # Apply FF correction - Manual
    choice4.addItemListener(previewer2)  # Apply FF correction - Manual

    gd.showDialog()

    #Get Parameters
    parm_Mode   = "Auto"
    TRCselected   = gd.getNextChoice()
    if TRCselected == "Perceptual":
        parm_FGGain = 128
    else:
        parm_FGGain = 56
    parm_Binary   = gd.getNextNumber() # Fg / Bg threshold
    parm_Gain = gd.getNextNumber() # (5.01a) Gamma value for Bg
    parm_Shadow = gd.getNextNumber() # Shadow correction threshold
    parm_Prph = gd.getNextNumber() # Peripheral fading correction threshold
    parm_HybridC   = gd.getNextChoice() # Correction method
    if parm_HybridC == "New Hybrid":
        parm_HybridC = "New"
        parm_FFCorrection = False
    elif parm_HybridC == "G Mixing":
        parm_HybridC = "Off"
        parm_FFCorrection = False
    elif parm_HybridC == "Hybrid":
        parm_HybridC = "On"
        parm_FFCorrection = False
    elif parm_HybridC == "EPFFc":
        parm_HybridC = "Off"
        parm_FFCorrection = True
    Fg_mask_type  = gd.getNextChoice() # Fg layer mask type
#    parm_FFCorrection = gd.getNextBoolean() #get value of checkbox
    parm_FFC_sigma = gd.getNextNumber()
                # Sigma for Gaussian filtering for psedo FF correction
    parm_Gmixrate = gd.getNextNumber() # G mixing rate for B
    parm_leveladj  = gd.getNextChoice() # apply level adjustment between channels?
    parm_correct_fac = gd.getNextNumber() # correction adjustment factor
    parm_subtractBv = gd.getNextBoolean() #get value of checkbox
    parm_ReadMask = gd.getNextBoolean() # read user made mask?
    DoPremixing = gd.getNextBoolean() # apply premixing for original B image?
    parm_Condition   = gd.getNextChoice() # Image condition Normal / Coarse

    parm_Show   = gd.getNextChoice() # show correcting process?
    bgG_gamma = 0.69
    if parm_Binary > 255: #Set Max Value
        parm_Binary = 255
    if parm_Gain > 1.7:
        parm_Gain = 1.7
    bgG_gamma = 1.0 / parm_Gain
    parm_Gain = 0.0
    if parm_Shadow > 255: #Set Max Value
        parm_Shadow = 255
        if parm_Gmixrate > 1.0:
            parm_Gmixrate = 1.0
        elif parm_Gmixrate < 0.0:
            parm_Gmixrate = 0.0
    if gd.wasCanceled():
        parm_Mode   = "Stop"
        IJ.log("Canceled")
        return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow,\
            str(parm_Show), parm_Prph, parm_FGGain, parm_FFCorrection, \
            parm_FFC_sigma, parm_Gmixrate, parm_subtractBv, parm_leveladj, \
            parm_correct_fac, parm_HybridC, Fg_mask_type, \
            bgG_gamma, parm_ReadMask, DoPremixing
    write_out(path, parm_Binary, parm_Gain, parm_Shadow, parm_Prph, \
            parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, \
            Fg_mask_type, bgG_gamma, parm_Condition)
    return str(parm_Mode), parm_Binary, parm_Gain, parm_Shadow, \
        str(parm_Show), parm_Prph, parm_FGGain, parm_FFCorrection, \
        parm_FFC_sigma, parm_Gmixrate, parm_subtractBv, parm_leveladj, \
        parm_correct_fac, parm_HybridC, Fg_mask_type, bgG_gamma, \
        parm_ReadMask, DoPremixing
#----------------
def write_out(path, parm_Binary, parm_Gain, parm_Shadow, parm_Prph, \
            parm_FGGain, parm_FFCorrection, parm_FFC_sigma, parm_Gmixrate, \
            parm_subtractBv, parm_leveladj, parm_correct_fac, parm_HybridC, \
            Fg_mask_type, bgG_gamma, parm_Condition):
    wo_parm = dict(parm_Binary=parm_Binary, parm_Gain=parm_Gain, \
        parm_Shadow=parm_Shadow, parm_Prph=parm_Prph, \
        parm_FGGain=parm_FGGain, parm_FFCorrection=parm_FFCorrection, \
        parm_FFC_sigma=parm_FFC_sigma, parm_Gmixrate=parm_Gmixrate, \
        parm_subtractBv=parm_subtractBv, parm_leveladj=parm_leveladj, \
        parm_correct_fac=parm_correct_fac, parm_HybridC=parm_HybridC, \
        Fg_mask_type=Fg_mask_type, bgG_gamma=bgG_gamma, \
        parm_Condition=parm_Condition)
#  -- writing "filename_parameter.txt" file
    writing_file0 = open(path  + '_parameter.txt','a')
    # new_file_number = int(file_number) + 1
    # i = 0
    for key in wo_parm.keys():
        writing_file0.write(key + ": " + str(wo_parm[key]) +"\n")
    timenow = datetime.datetime.now()
    writing_file0.write("date time: " + \
        timenow.strftime('%Y/%m/%d/%H:%M:%S') +"\n")
    writing_file0.close()
#-------------
def read_in(path, Vmean):
    existing_parms = {}
        # define dic type Variable
    # set default value
    existing_parms["parm_FFC_sigma"] = 10.0
    existing_parms["parm_subtractBv"] = "False"
    existing_parms["parm_leveladj"] = "Auto"
    existing_parms["parm_Shadow"] = 25.0
    existing_parms["parm_Prph"] = 10.0
    existing_parms["parm_correct_fac"] = 0.0
    if Vmean < 128.0:
        BkThs = 128.0
    else:
        BkThs = int(Vmean)

    existing_parms["parm_Binary"] = BkThs
    existing_parms["parm_Gmixrate"] = 0.0
    existing_parms["parm_Gain"] = 0.0
    existing_parms["parm_FFCorrection"] = "False"
    existing_parms["parm_FGGain"] = 128
    existing_parms["parm_HybridC"] = "Off"
    existing_parms["Fg_mask_type"] = "Type2"
    existing_parms["bgG_gamma"] = 0.64
    existing_parms["parm_Condition"] = "Normal"
    existing_parms["existing"] = "No"
    is_file = os.path.isfile(path + '_parameter.txt')
    if is_file:
        # reading existing parameters
        with open(path + '_parameter.txt','r') as read_file:
            lines = read_file.readlines()
            if len(lines) >= 15:
                for l in range(len(lines)-15, len(lines)-1):
                    eachline = lines[l].strip()
                    i = eachline.find(":")
                    key = eachline[:i]
                    value = eachline[i+2:]
                    existing_parms[key] = value
                    existing_parms["existing"] = "Yes"
    # print "existing: ", existing_parms["existing"], "Lines: ", len(lines)

    return existing_parms
#-------------
# Main
JobDo = True
window = WM.getActiveWindow()
imgCurrent = WM.getCurrentImage()
if not imgCurrent:
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    file = od.getFileName()
    if file:
        path = folder + file
        IJ.log(path)
        IJ.log(folder)
        IJ.log(file)
        imps = BF.openImagePlus(path) # Open file with Bio-format Plugin
        imp0 = imps[0]
    else:
        IJ.log("Job Canceled")
        JobDo = False
else: # target image is already open
    impgot = IJ.getImage()
    imps = [impgot]
    file = imps[0].getTitle()
    try:
        folder = imps[0].getOriginalFileInfo().directory
    except:
        folder = IJ.getDirectory("Output_directory")
    path = folder + file
if JobDo == True:
    bitdepth = imps[0].getBitDepth()
    if bitdepth != 0:
            if bitdepth == 32:
                L_max = set_Lmax(imps[0], "32-bit")
            stat0 = imps[0].getStatistics()
            Vmean = stat0.mean
            if bitdepth == 16:
                Vmean = int(Vmean / 256)
            elif bitdepth ==32:
                Vmean = int(convValue1(Vmean, L_max))
            Vmean = Vmean + 3
            if bitdepth == 32:
                IJ.showMessage("32bit Max L Value: " + str(L_max))
            parm = NoviceMenu(path, Vmean)
            if parm[0] == "Expert":
                parm = getParameters(path, Vmean)
            print ("Parm[0]=", parm[0])
            if parm[0] != "Stop":
                RGBmanipulation(folder,file,parm, imps)
    else:
        IJ.showMessage("This image is not compatible with this program.")
IJ.log("Finish")
#    IJ.run("Quit")
